import math


def optimal_bits(value_range: 'tuple[int, int]') -> int:
    """Returns the optimal number of bits for encoding a specified range.
    
    Args:
        value_range: A tuple with the minimum and maximum values.
    
    Returns:
        The number of bits to optimally encode the value.
    
    Raises:
        ValueError if the 

    """
    if (not isinstance(value_range, tuple) or
        len(value_range) != 2 or
        not all(isinstance(x, int) for x in value_range) or
        value_range[0] >= value_range[1]):
        #: non-compliant
        raise ValueError('value_range must be of form (min, max)')
    total_range = value_range[1] - value_range[0]
    total_range += 1 if value_range[0] == 0 else 0
    return max(1, math.ceil(math.log2(total_range)))


def encode_field_length(length) -> str:
    if length < 128:
        return f'0{length:07b}'
    return f'1{length:015b}'


def decode_field_length(binstr: str) -> 'tuple[int, int]':
    if binstr[0] == '0':
        bit_index = 8
    else:
        bit_index = 16
    length = int(binstr[1:bit_index], 2)
    return (length, bit_index)



from warnings import warn

from . import ET
from .base import BaseCodec, CodecList
from .messages import MessageCodec, Messages


class ServiceCodec(BaseCodec):
    """A data structure holding a set of related Forward and Return Messages.
    
    Attributes:
        name (str): The service name
        sin (int): Service Identification Number or codec service id (16..255)
        description (str): A description of the service (unsupported)
        messages_forward (list): A list of mobile-terminated Message definitions
        messages_return (list): A list of mobile-originated Message definitions

    """
    def __init__(self,
                 name: str,
                 sin: int,
                 description: str = None,
                 messages_forward: Messages = None,
                 messages_return: Messages = None) -> None:
        """Instantiates a Service made up of Messages.
        
        Args:
            name: The service name should be unique within a MessageDefinitions
            sin: The Service Identification Number (16..255)
            description: (Optional)
        """
        if not isinstance(name, str) or name == '':
            raise ValueError(f'Invalid service name {name}')
        if sin not in range(16, 256):
            raise ValueError('Invalid SIN must be 16..255')
        if description is not None:
            warn('Service Description not currently supported')
        super().__init__(name, description)
        self._sin = sin
        self._messages_forward = (messages_forward or
                                  Messages(self.sin, is_forward=True))
        self._messages_return = (messages_return or
                                 Messages(self.sin, is_forward=False))
    
    @property
    def sin(self) -> int:
        return self._sin
    
    @property
    def messages_forward(self) -> Messages:
        return self._messages_forward
    
    @messages_forward.setter
    def messages_forward(self, messages: Messages):
        if not isinstance(messages, Messages):
            raise ValueError('Invalid messages list')
        for message in messages:
            assert isinstance(message, MessageCodec)
            if not message.is_forward:
                raise ValueError(f'Message {message.name} is_forward is False')
        self._messages_forward = messages

    @property
    def messages_return(self) -> Messages:
        return self._messages_return
    
    @messages_return.setter
    def messages_return(self, messages: Messages):
        if not isinstance(messages, Messages):
            raise ValueError('Invalid messages list')
        for message in messages:
            assert isinstance(message, MessageCodec)
            if message.is_forward:
                raise ValueError(f'Message {message.name} is_forward is True')
        self._messages_return = messages
        
    def xml(self) -> ET.Element:
        """Returns the Service XML definition for a Message Definition File."""
        if len(self.messages_forward) == 0 and len(self.messages_return) == 0:
            raise ValueError(f'No messages defined for service {self.sin}')
        xservice = ET.Element('Service')
        name = ET.SubElement(xservice, 'Name')
        name.text = str(self.name)
        sin = ET.SubElement(xservice, 'SIN')
        sin.text = str(self.sin)
        if self.description:
            desc = ET.SubElement(xservice, 'Description')
            desc.text = str(self.description)
        if len(self.messages_forward) > 0:
            forward_messages = ET.SubElement(xservice, 'ForwardMessages')
            for m in self.messages_forward:
                forward_messages.append(m.xml())
        if len(self.messages_return) > 0:
            return_messages = ET.SubElement(xservice, 'ReturnMessages')
            for m in self.messages_return:
                return_messages.append(m.xml())
        return xservice


class Services(CodecList):
    """The list of Service(s) within a MessageDefinitions."""
    def __init__(self, services: 'list[ServiceCodec]' = None):
        super().__init__(codec_cls=ServiceCodec)
        if services is not None:
            for service in services:
                if not isinstance(service, ServiceCodec):
                    raise ValueError(f'Invalid Service {service}')
                self.add(service)
    
    def add(self, service: ServiceCodec) -> None:
        """Adds a Service to the list of Services."""
        if not isinstance(service, ServiceCodec):
            raise ValueError(f'{service} is not a valid Service')
        if service.name in self:
            raise ValueError(f'Duplicate Service {service.name}')
        for existing_service in self:
            if existing_service.sin == service.sin:
                raise ValueError(f'Duplicate SIN {service.sin}')
        self.append(service)



from struct import pack, unpack
from warnings import warn

from .. import ET
from .base_field import FieldCodec
from .helpers import decode_field_length, encode_field_length


class DataField(FieldCodec):
    """A data field of raw bytes sent over-the-air.
    
    Can also be used to hold floating point, double-precision or large integers.

    """
    SUPPORTED_DATA_TYPES = ['data', 'float', 'double']
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'data',
                 precision: int = None,
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 default: bytes = None,
                 value: bytes = None) -> None:
        """Instantiates a EnumField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The maximum number of bytes to send over-the-air.
            data_type: The data type represented within the bytes.
            precision: The number of decimal places for float/double.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            fixed: Indicates if the data bytes are a fixed `size`.
            default: A default value for the enum.
            value: Optional value to set during initialization.

        """
        if data_type is None or data_type not in self.SUPPORTED_DATA_TYPES:
            raise ValueError(f'Invalid data type {data_type}')
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self._fixed = fixed
        self._size = None
        self._default = None
        self._precision = None
        self._value = None
        self.precision = precision
        self.size = size
        self.default = default
        self.value = value
    
    @property
    def size(self) -> int:
        """The maximum size of the field in bytes."""
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bytes')
        if self.data_type == 'float':
            if value != 4:
                warn('Adjusting float size to 4 bytes fixed')
            self._size = 4
            self._fixed = True
        elif self.data_type == 'double':
            if value != 8:
                warn('Adjusting double size to 8 bytes fixed')
            self._size = 8
            self._fixed = True
        else:
            self._size = value
    
    def _validate_data(self, v: 'bytes|float') -> bytes:
        """Ensures the data is of the target field size and encoding."""
        data_type = self.data_type
        if not ((isinstance(v, bytes) and data_type == 'data') or
                (isinstance(v, float) and data_type in ['float', 'double'])):
            raise ValueError(f'Data {type(v)} does not match {data_type}')
        if data_type in ['float', 'double']:
            _format = '!f' if data_type == 'float' else '!d'
            v = pack(_format, v)
        assert isinstance(v, bytes)
        if self.fixed:
            if len(v) > self.size:
                warn(f'Truncating data to {self.size} bytes')
                return v[0:self.size]
            elif len(v) < self.size:
                warn(f'Padding data to {self.size} bytes')
                return v.ljust(self.size, b'\0')
        return v

    def _convert_to_float(self, v: bytes) -> 'float|None':
        if self.data_type not in ('float', 'double') or v is None:
            return None
        convertor = '!f' if self.data_type == 'float' else '!d'
        converted = unpack(convertor, v)[0]
        if self.precision:
            converted = round(converted, self.precision)
        return converted

    @property
    def default(self) -> 'bytes|float':
        """The default value, converted for float or double data types."""
        if self.data_type in ['float', 'double']:
            return self._convert_to_float(self._default)
        return self._default
    
    @default.setter
    def default(self, v: 'bytes|float'):
        if v is None:
            self._default = None
        else:
            self._default = self._validate_data(v)

    @property
    def precision(self) -> 'int|None':
        """The number of decimal places for `float` or `double` data types."""
        return self._precision
    
    @precision.setter
    def precision(self, value: 'int|None'):
        if self.data_type in ['float', 'double']:
            if value is not None and not isinstance(value, int):
                raise ValueError('Precision must be int or None'
                                 ' for float/double data_type')
        elif value is not None:
            raise ValueError('Precision only valid for float/double data_type')
        self._precision = value

    @property
    def converted_value(self) -> 'float|None':
        """The converted value for `float` and `double` data types."""
        return self._convert_to_float(self._value)
    
    @property
    def value(self):
        """The raw binary value."""
        return self._value

    @value.setter
    def value(self, v: 'bytes|float'):
        if v is None:
            self._value = None
        else:
            self._value = self._validate_data(v)

    @property
    def fixed(self) -> bool:
        """Indicates if the field is fixed size (padded/truncated)."""
        return self._fixed

    @property
    def bits(self):
        """The size of the field in bits."""
        if self._value is None:
            bits = 0
        elif self.fixed:
            bits = self.size * 8
        else:
            L = 8 if len(self._value) < 127 else 16
            bits = L + len(self._value) * 8
        return bits + (1 if self.optional else 0)
    
    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError(f'No value defined for DataField {self.name}')
        binstr = ''
        binstr = ''.join(format(b, '08b') for b in self._value)
        if self.fixed:   #:pad to fixed length
            binstr += ''.join('0' for bit in range(len(binstr), self.bits))
        else:
            binstr = encode_field_length(len(self._value)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            binary = binary_str[:self.bits]
            bits = self.bits
        else:
            (length, bit_index) = decode_field_length(binary_str)
            binary = binary_str[bit_index:length * 8 + bit_index]
            bits = len(binary)
        self._value = int(binary, 2).to_bytes(int(bits / 8), 'big')
        return self.bits

    def xml(self) -> ET.Element:
        """Returns the Data XML definition for a Message Definition File."""
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield



import xml.etree.ElementTree as ET

from .base import BaseCodec, CodecList
from .constants import *
from .fields import (ArrayField, BooleanField, DataField, EnumField,
                     SignedIntField, StringField, UnsignedIntField)
from .fields.base_field import FieldCodec, Fields
from .fields.helpers import optimal_bits
from .message_definitions import MessageDefinitions
from .messages import MessageCodec, Messages
from .services import ServiceCodec, Services

for ns in XML_NAMESPACE:
    ET.register_namespace(ns, XML_NAMESPACE[ns])



import math
from binascii import b2a_base64

from . import ET
from .base import BaseCodec, CodecList
from .fields.base_field import FieldCodec, Fields

from idpmodem.constants import DataFormat


class MessageCodec(BaseCodec):
    """The Payload structure for Message Definition Files uploaded to a Mailbox.
    
    Attributes:
        name (str): The message name
        sin (int): The Service Identification Number
        min (int): The Message Identification Number
        fields (list): A list of Fields
        description (str): Optional description
        is_forward (bool): Indicates if the message is mobile-terminated

    """

    def __init__(self,
                 name: str,
                 sin: int,
                 min: int,
                 description: str = None,
                 is_forward: bool = False,
                 fields: Fields = None):
        """Instantiates a Message.
        
        Args:
            name: The message name should be unique within the xMessages list.
            sin: The Service Identification Number (16..255)
            min: The Message Identification Number (0..255)
            description: (Optional) Description/purpose of the Message.
            is_forward: Indicates if the message is intended to be
                Mobile-Terminated.
            fields: Optional definition of fields during instantiation.

        """
        if not isinstance(sin, int) or sin not in range(16, 256):
            raise ValueError(f'Invalid SIN {sin} must be in range 16..255')
        if not isinstance(min, int) or min not in range (0, 256):
            raise ValueError(f'Invalid MIN {min} must be in range 0..255')
        super().__init__(name, description)
        self._is_forward = is_forward
        self._sin = sin
        self._min = min
        self._fields: Fields = fields or Fields()

    @property
    def is_forward(self) -> bool:
        return self._is_forward
    
    @property
    def sin(self) -> int:
        return self._sin

    @property
    def min(self) -> int:
        return self._min

    @property
    def fields(self) -> Fields:
        return self._fields
    
    @fields.setter
    def fields(self, fields: Fields):
        if not all(isinstance(field, FieldCodec) for field in fields):
            raise ValueError('Invalid field found in list')
        self._fields = fields

    @property
    def ota_size(self) -> int:
        ota_bits = 2 * 8
        for field in self.fields:
            assert isinstance(field, FieldCodec)
            ota_bits += field.bits
        return math.ceil(ota_bits / 8)

    def decode(self, databytes: bytes) -> None:
        """Parses and stores field values from raw data (received over-the-air).
        
        Args:
            databytes: A bytes array (typically from the forward message)
        """
        binary_str = ''.join(format(int(b), '08b') for b in databytes)
        bit_offset = 16   #: Begin after SIN/MIN bytes
        for field in self.fields:
            assert isinstance(field, FieldCodec)
            if field.optional:
                present = binary_str[bit_offset] == '1'
                bit_offset += 1
                if not present:
                    continue
            bit_offset += field.decode(binary_str[bit_offset:])

    def encode(self,
               data_format: int = DataFormat.BASE64,
               exclude: list = None) -> dict:
        """Encodes using the specified data format (base64 or hex).

        Args:
            data_format (int): 2=ASCII-Hex, 3=base64
            exclude (list[str]): A list of optional field names to exclude
        
        Returns:
            Dictionary with sin, min, data_format and data to pass into AT%MGRT
                or atcommand function `message_mo_send`
        """
        if data_format not in [DataFormat.BASE64, DataFormat.HEX]:
            raise ValueError(f'data_format {data_format} unsupported')
        bin_str = ''
        for field in self.fields:
            assert isinstance(field, FieldCodec)
            if field.optional:
                if exclude is not None and field.name in exclude:
                    present = False
                elif hasattr(field, 'value'):
                    present = field.value is not None
                elif hasattr(field, 'elements'):
                    present = field.elements is not None
                else:
                    raise ValueError('Unknown value of optional')
                bin_str += '1' if present else '0'
                if not present:
                    continue
            bin_str += field.encode()
        for _ in range(0, 8 - len(bin_str) % 8):   #:pad to next byte
            bin_str += '0'
        _format = f'0{int(len(bin_str) / 8 * 2)}X'   #:hex bytes 2 chars
        hex_str = format(int(bin_str, 2), _format)
        if (self.is_forward and len(hex_str) / 2 > 9998 or
            not self.is_forward and len(hex_str) / 2 > 6398):
            raise ValueError(f'{len(hex_str) / 2} bytes exceeds maximum size'
                             ' for Payload')
        if data_format == DataFormat.HEX:
            data = hex_str
        else:
            data = b2a_base64(bytearray.fromhex(hex_str)).strip().decode()
        return {
            'sin': self.sin,
            'min': self.min,
            'data_format': data_format,
            'data': data
        }

    def xml(self) -> ET.Element:
        """Returns the Message XML definition for a Message Definition File."""
        xmessage = ET.Element('Message')
        name = ET.SubElement(xmessage, 'Name')
        name.text = self.name
        min = ET.SubElement(xmessage, 'MIN')
        min.text = str(self.min)
        fields = ET.SubElement(xmessage, 'Fields')
        for field in self.fields:
            fields.append(field.xml())
        return xmessage


class Messages(CodecList):
    """The list of Messages (Forward or Return) within a Service."""
    def __init__(self, sin: int, is_forward: bool):
        super().__init__(codec_cls=MessageCodec)
        self.sin = sin
        self.is_forward = is_forward
    
    def add(self, message: MessageCodec) -> None:
        """Add a message to the list if it matches the parent SIN.

        Overrides the base class add method.

        Args:
            message (object): A valid Message
        
        Raises:
            ValueError if there is a duplicate or invalid name,
                invalid value_range or unsupported data_type

        """
        if not isinstance(message, MessageCodec):
            raise ValueError('Invalid message definition')
        if message.sin != self.sin:
            raise ValueError(f'Message SIN {message.sin} does not match'
                             f' service {self.sin}')
        for m in self:
            assert isinstance(m, MessageCodec)
            if m.name == message.name:
                raise ValueError(f'Duplicate message name {message.name} found')
            if m.min == message.min:
                raise ValueError(f'Duplicate message MIN {message.min} found')
        self.append(message)



from warnings import warn

from .. import ET
from .base_field import FieldCodec
from .helpers import decode_field_length, encode_field_length


class StringField(FieldCodec):
    """A character string sent over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 default: str = None,
                 value: str = None) -> None:
        """Instantiates a StringField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The maximum number of characters in the string.
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            fixed: Indicates if the string is always fixed length `size`.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        super().__init__(name=name,
                         data_type='string',
                         description=description,
                         optional=optional)
        self._size = size
        self._fixed = fixed
        self._default = default
        self._value = value if value is not None else self._default
    
    def _validate_string(self, s: str) -> str:
        if s is not None:
            if not isinstance(s, str):
                raise ValueError(f'Invalid string {s}')
            if len(s) > self.size:
                warn(f'Clipping string at max {self.size} characters')
                return s[:self.size]
        return s
                
    @property
    def size(self) -> int:
        """The maximum size of the string in characters."""
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 characters')
        self._size = value
    
    @property
    def default(self) -> str:
        """The default value."""
        return self._default
    
    @default.setter
    def default(self, v: str):
        self._default = self._validate_string(v)

    @property
    def value(self) -> str:
        return self._value
    
    @value.setter
    def value(self, v: str):
        self._value = self._validate_string(v)

    @property
    def fixed(self) -> bool:
        """Indicates whether the string length is fixed (padded/truncated)."""
        return self._fixed
    
    @fixed.setter
    def fixed(self, value: bool):
        self._fixed = value

    @property
    def bits(self) -> int:
        """The size of the field in bits."""
        if self._value is None:
            bits = 0
        elif self.fixed:
            bits = self.size * 8
        else:
            L = 8 if len(self._value) < 127 else 16
            bits = L + len(self._value) * 8
        return bits + (1 if self.optional else 0)
    
    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError(f'No value defined for StringField {self.name}')
        binstr = ''.join(format(ord(c), '08b') for c in self.value)
        if self.fixed:
            binstr += ''.join('0' for bit in range(len(binstr), self.bits))
        else:
            binstr = encode_field_length(len(self.value)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            length = self.size
            bit_index = 0
        else:
            (length, bit_index) = decode_field_length(binary_str)
        n = int(binary_str[bit_index:bit_index + length * 8], 2)
        char_bytes = n.to_bytes((n.bit_length() + 7) // 8, 'big')
        for i, byte in enumerate(char_bytes):
            if byte == 0:
                warn('Truncating after 0 byte in string')
                char_bytes = char_bytes[:i]
                break
        self.value = char_bytes.decode('utf-8', 'surrogatepass') or '\0'
        return bit_index + length * 8

    def xml(self) -> ET.Element:
        """Returns the String XML definition for a Message Definition File."""
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.fixed:
            fixed = ET.SubElement(xmlfield, 'Fixed')
            fixed.text = 'true'
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield



from .. import ET
from .base_field import FieldCodec


class BooleanField(FieldCodec):
    """A Boolean field."""
    def __init__(self,
                 name: str,
                 description: str = None,
                 optional: bool = False,
                 default: bool = False,
                 value: bool = None) -> None:
        super().__init__(name=name,
                         data_type='bool',
                         description=description,
                         optional=optional)
        """Instantiates a BooleanField.
        
        Args:
            name: The field name must be unique within a Message.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            default: A default value for the boolean.
            value: Optional value to set during initialization.

        """
        self._default = default if isinstance(default, bool) else False
        self._value = value if value is not None else self._default
    
    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, v: bool):
        if v is not None and not isinstance(v, bool):
            raise ValueError(f'Invalid boolean value {v}')
        self._default = v

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: bool):
        if v is not None and not isinstance(v, bool):
            raise ValueError(f'Invalid boolean value {v}')
        self._value = v

    @property
    def bits(self):
        bits = 0 if self._value is None else 1
        return bits + (1 if self.optional else 0)
    
    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError('No value assigned to field')
        return '1' if self.value else '0'

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = True if binary_str[0] == '1' else False
        return 1

    def xml(self) -> ET.Element:
        """Returns the Boolean XML definition for a Message Definition File."""
        xmlfield = self._base_xml()
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = 'true'
        return xmlfield



from copy import deepcopy

from . import ET, XML_NAMESPACE
from .services import ServiceCodec, Services


class MessageDefinitions:
    """A set of Message Definitions grouped into Services.

    Attributes:
        services: The list of Services with Messages defined.
    
    """
    def __init__(self, services: Services = None):
        if services is not None:
            if not isinstance(services, Services):
                raise ValueError('Invalid Services')
        self.services = services or Services()
    
    def xml(self) -> ET.Element:
        """Gets the XML structure of the complete message definitions."""
        xmsgdef = ET.Element('MessageDefinition',
                             attrib={'xmlns:xsd': XML_NAMESPACE['xsd']})
        services = ET.SubElement(xmsgdef, 'Services')
        for service in self.services:
            assert isinstance(service, ServiceCodec)
            services.append(service.xml())
        return xmsgdef
    
    def mdf_export(self,
                   filename: str,
                   pretty: bool = False,
                   indent: int = 0,
                   include_service_description: bool = False,
                   ) -> None:
        """Creates an XML file at the target location.
        
        Args:
            filename: The full path/filename to save to. `.idpmsg` is
                recommended as a file extension.
            pretty: If True sets indent = 2 (legacy compatibility)
            indent: If nonzero will indent each layer of the XML by n spaces.
            include_service_description: By default removes Description from
                Service for Inmarsat IDP Admin API V1 compatibility.

        """
        if not include_service_description:
            new_copy = deepcopy(self)
            for service in new_copy.services:
                assert isinstance(service, ServiceCodec)
                if service.description is not None:
                    service.description = None
            tree = ET.ElementTree(new_copy.xml())
        else:
            tree = ET.ElementTree(self.xml())
        if pretty:
            indent = 2
        if indent:
            root = tree.getroot()
            _indent(root, spaces=indent)
            xmlstr = ET.tostring(root).decode()
            with open(filename, 'w') as f:
                f.write(xmlstr)
        else:
            with open(filename, 'wb') as f:
                tree.write(f, encoding='utf-8', xml_declaration=True)


def _indent(elem: ET.Element, level: int = 0, spaces: int = 2) -> ET.Element:
    i = '\n' + level * (' ' * spaces)
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + ' ' * spaces
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        sub_index = 0
        for subelem in elem:
            sub_index += 1
            _indent(subelem, level + 1)
            if sub_index == len(elem):
                subelem.tail = i
        # if not elem.tail or not elem.tail.strip():
        #     elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i
    if level == 0:
        elem.tail = None



from warnings import warn

from .. import ET
from .base_field import FieldCodec


class UnsignedIntField(FieldCodec):
    """An unsigned integer value using a defined number of bits over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'uint_16',
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a UnsignedIntField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The number of *bits* used to encode the integer over-the-air
                (maximum 32).
            data_type: The integer type represented (for decoding).
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        if data_type not in ['uint_8', 'uint_16', 'uint_32']:
            raise ValueError(f'Invalid unsignedint type {data_type}')
        if not isinstance(size, int) or size < 1:
            raise ValueError('Size must be int greater than zero')
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self._size = size
        self._default = default
        self._value = value if value is not None else self._default
    
    @property
    def size(self):
        """The size of the field in bits."""
        return self._size

    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        data_type_size = int(self.data_type.split('_')[1])
        if value > data_type_size:
            warn(f'Size {value} larger than required by {self.data_type}')
        self._size = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: int):
        clip = False
        if v is not None:
            if not isinstance(v, int) or v < 0:
                raise ValueError('Unsignedint must be non-negative integer')
            if v > 2**self.size - 1:
                self._value = 2**self.size - 1
                warn(f'Clipping unsignedint at max value {self._value}')
                clip = True
        if not clip:
            self._value = v
    
    @property
    def default(self):
        """The default value."""
        return self._default
    
    @default.setter
    def default(self, v: int):
        if v is not None:
            if v > 2**self.size - 1 or v < 0:
                raise ValueError(F'Invalid unsignedint default {v}')
        self._default = v
    
    @property
    def bits(self):
        """The size of the field in bits."""
        bits = self.size if self._value is not None else 0
        return bits + (1 if self.optional else 0)
    
    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError(f'No value defined in UnsignedIntField {self.name}')
        _format = f'0{self.size}b'
        return format(self.value, _format)

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = int(binary_str[:self.size], 2)
        return self.size

    def xml(self) -> ET.Element:
        """Returns the UnsignedInt XML definition for a Message Definition File.
        """
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class SignedIntField(FieldCodec):
    """A signed integer value using a defined number of bits over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'int_16',
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a SignedIntField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The number of *bits* used to encode the integer over-the-air
                (maximum 32).
            data_type: The integer type represented (for decoding).
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        if data_type not in ['int_8', 'int_16', 'int_32']:
            raise ValueError(f'Invalid unsignedint type {data_type}')
        if not isinstance(size, int) or size < 1:
            raise ValueError('Size must be int greater than zero')
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self._size = size
        self._default = default
        self._value = value if value is not None else self._default
    
    @property
    def size(self):
        """The size of the field in bits."""
        return self._size

    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        self._size = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: int):
        clip = False
        if v is not None:
            if not isinstance(v, int):
                raise ValueError('Unsignedint must be non-negative integer')
            if v > (2**self.size / 2) - 1:
                self._value = int(2**self.size / 2) - 1
                warn(f'Clipping signedint at max value {self._value}')
                clip = True
            if v < -(2**self.size / 2):
                self._value = -1 * int(2**self.size / 2)
                warn(f'Clipping signedint at min value {self._value}')
                clip = True
        if not clip:
            self._value = v
    
    @property
    def default(self):
        """The default value."""
        return self._default
    
    @default.setter
    def default(self, v: int):
        if v is not None:
            if not isinstance(v, int):
                raise ValueError(f'Invalid signed integer {v}')
            if v > (2**self.size / 2) - 1 or v < -(2**self.size / 2):
                raise ValueError(f'Invalid default {v}')
        self._default = v
    
    @property
    def bits(self):
        """The size of the field in bits."""
        bits = self.size if self._value is not None else 0
        return bits + (1 if self.optional else 0)
    
    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError(f'No value defined in UnsignedIntField {self.name}')
        _format = f'0{self.size}b'
        if self.value < 0:
            invertedbin = format(self.value * -1, _format)
            twocomplementbin = ''
            i = 0
            while len(twocomplementbin) < len(invertedbin):
                twocomplementbin += '1' if invertedbin[i] == '0' else '0'
                i += 1
            binstr = format(int(twocomplementbin, 2) + 1, _format)
        else:
            binstr = format(self.value, _format)
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        value = int(binary_str[:self.size], 2)
        if (value & (1 << (self.size - 1))) != 0:   #:sign bit set e.g. 8bit: 128-255
            value = value - (1 << self.size)        #:compute negative value
        self.value = value
        return self.size

    def xml(self) -> ET.Element:
        """Returns the SignedInt XML definition for a Message Definition File.
        """
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield
